from typing import Any, Sequence, Tuple, Dict, Optional

import numpy as np

from bokeh.layouts import column
from bokeh.models import Select, Slider, RangeSlider, ColorMapper
from bokeh.models.formatters import BasicTickFormatter
from pyscivis.visualizer.dataclasses.state import PaletteState

Palette = Tuple[str]


class PaletteControls:
    """
    Wraps all palette-related controls and their callbacks.
    """

    def __init__(self,
                 color_mappers: Sequence[ColorMapper],
                 palette_map: Dict[str, Sequence[Palette]],
                 palette_size: int,
                 initial_palette: str,
                 min_: float,
                 max_: float,
                 **kwargs: Any
                 ) -> None:
        """
        Set up all controls and their callbacks.

        Args:
            color_mappers: Sequence of ColorMapper to be manipulated.
            palette_map: Dictionary mapping palette names to Sequence of palettes.
            palette_size: Maximal palette size.
            initial_palette: Name of the initial palette.
            min_: Minimum of data.
            max_: Maximum of data.
            **kwargs: Keywords passed onto the LayoutDOM containing the controls.
        """
        same_min_max_flag = False
        if min_ == max_ or np.isnan(min_) or np.isnan(max_):
            same_min_max_flag = True
            min_ = 0
            max_ = 1

        sorted_keys = sorted(palette_map.keys())

        self.palette = Select(title='Palette', value=initial_palette,
                              options=sorted_keys, sizing_mode="stretch_width",
                              css_classes=["palette-select"])

        self.palette_length = Slider(title="Palette-size", start=1,
                                     end=palette_size, value=palette_size, step=1, sizing_mode="stretch_width",
                                     css_classes=["palette-size"])

        self.palette_window = RangeSlider(title="Palette-window", start=min_, end=max_, format=BasicTickFormatter(),
                                          value=(min_, max_), step=0, sizing_mode="stretch_width",
                                          css_classes=["palette-window"], disabled=True)

        self.palette_cutoff = RangeSlider(title="Window", start=min_, end=max_, format=BasicTickFormatter(),
                                          value=(min_, max_), step=0, sizing_mode="stretch_width",
                                          css_classes=["palette-slider"], disabled=True)

        self.layout = column(self.palette, self.palette_length, self.palette_window, self.palette_cutoff,
                             **kwargs, css_classes=["palette"])

        def _palette_cb(attr: Any,
                        old: Any,
                        new: Any
                        ) -> None:
            """Change palette on name, size and cutoff change."""
            # select palette kind
            full_palette = list(palette_map[self.palette.value][self.palette_length.value-1])
            # calculate the cutoff-window
            dist = self.palette_cutoff.end - self.palette_cutoff.start
            dist_front = self.palette_cutoff.value[0] - self.palette_cutoff.start
            dist_back = self.palette_cutoff.end - self.palette_cutoff.value[1]
            ratio_front_total = dist_front/dist  # ratio of how much is to be cut off in the front
            ratio_back_total = dist_back/dist  # ratio of how much is to be cut off in the back
            front_color = full_palette[0]  # used to pad everything in front of the first slider-button
            back_color = full_palette[-1]  # used to pad everything behind the second slider button
            front_colorize_length = round(ratio_front_total * len(full_palette))
            back_colorize_length = round(ratio_back_total * len(full_palette))
            # now overwrite the palette's front and back cutoff
            full_palette[:front_colorize_length] = [front_color for _ in full_palette[:front_colorize_length]]
            if back_colorize_length > 0:  # [-0:] would return entire array
                full_palette[-back_colorize_length:] = [back_color for _ in full_palette[-back_colorize_length:]]

            for cm in color_mappers:
                cm.palette = full_palette

        def _window_cb(attr: Any,
                       old: Any,
                       new: Tuple[float]
                       ) -> None:
            """Change palette on window change."""
            for cm in color_mappers:
                cm.update(low=new[0], high=new[1])

        self._palette_cb = _palette_cb
        self._window_cb = _window_cb

        self.palette.on_change('value', _palette_cb)
        self.palette_length.on_change('value', _palette_cb)

        if not same_min_max_flag:
            self.enable_and_attach_callbacks()

    def set_state(self, state: PaletteState):
        """
        Set state by updating the slider-positions of all controls.

        Args:
            state: PaletteState containing slider-values.
        """
        self.palette.value = state.name
        self.palette_length.value = state.length
        self.palette_window.value = state.window
        self.palette_cutoff.value = state.cutoff

    def get_state(self) -> PaletteState:
        """
        Get the state of the current PaletteControls.

        Returns:
            PaletteState containing slider-values.
        """
        plt: str = self.palette.value
        length: int = int(self.palette_length.value)
        plt_window: Tuple[float, float] = self.palette_window.value
        cutoff: Tuple[float, float] = self.palette_cutoff.value

        state = PaletteState(name=plt, length=length, window=plt_window, cutoff=cutoff)
        return state

    def reset_window_sliders(self,
                             min_: Optional[float] = None,
                             max_: Optional[float] = None
                             ) -> None:
        """
        Reset window and cutoff sliders.

        If min_ and max_ are specified the start and end values of window and cutoff are changed.
        This is typically necessary when the frame's minimum and maximum have changed.

        Always resets the current slider position.

        Args:
            min_: Minimum of data.
            max_: Maximum of data.
        """
        if min_ is not None and max_ is not None:
            if min_ == max_ or np.isnan(min_) or np.isnan(max_):
                self.disable_and_detach_callbacks()
                return
            self.enable_and_attach_callbacks()
            self.palette_window.update(start=min_, end=max_)
            self.palette_cutoff.update(start=min_, end=max_)

        self.palette_window.value = (self.palette_window.start, self.palette_window.end)
        self.palette_cutoff.value = (self.palette_cutoff.start, self.palette_cutoff.end)

    def enable_and_attach_callbacks(self):
        self.palette_cutoff.on_change('value', self._palette_cb)
        self.palette_window.on_change('value', self._window_cb)
        self.palette_window.disabled = False
        self.palette_cutoff.disabled = False

    def disable_and_detach_callbacks(self):
        self.palette_cutoff._callbacks.clear()
        self.palette_window._callbacks.clear()
        self.palette_window.disabled = True
        self.palette_cutoff.disabled = True



from bokeh.core.properties import Seq, Float
from bokeh.models import LogColorMapper


class OffsetLogColorMapper(LogColorMapper):
    """
    Allow logarithmic colormapping for negative intervals.

    Works exactly the same as its base class, however
    OffsetLogColorMapper also allows mapping datasets
    with negative values. This does not work like mpl's
    symlog but instead we use a standard natural log on
    an interval [1, 10] (unless otherwise specified), which
    the input data interval, e.g, [-5, 51] is then mapped to.

    See Also:
        OffsetLogTickFormatter
    """

    #__implementation__ = 'offset_log_color_mapper.ts'

    log_span = Seq(Float, default=[1, 10], help="""
            This is used to determine the steepness of
            the log-scale.
        """)


if __name__ == '__main__':
    import numpy as np
    from offset_log_tick_formatter import OffsetLogTickFormatter
    from bokeh.models import HoverTool
    from bokeh.palettes import turbo
    from bokeh.models import ColorBar
    from bokeh.plotting import figure
    from bokeh.io import show
    from bokeh.settings import settings
    settings.minified = False
    palette = turbo(256)
    #data_arr = [list(np.arange(-20, -10, 0.1)) for j in range(-10, 10)]
    #data_arr = [list(np.arange(-20, -1, 0.1)) for j in range(-10, 10)]
    data_arr = [list(np.arange(-505, 523, 1)) for j in range(-10, 10)]

    lg = OffsetLogColorMapper(palette)

    c = ColorBar(color_mapper=lg, formatter=OffsetLogTickFormatter(low=np.min(data_arr), high=np.max(data_arr)))
    fig = figure(width=500, height=500, tools=[])
    fig.image(image=[data_arr], x=0, y=0, dw=500, dh=500, color_mapper=lg)
    fig.add_tools(HoverTool(tooltips="@image"))
    fig.add_layout(c, "right")
    show(fig)



import threading
from typing import Any, Optional, Tuple, Union

from tornado import gen

from bokeh.io import curdoc
from bokeh.layouts import column, row
from bokeh.models import Spacer, Div, Button, ColumnDataSource, CustomJS, Dropdown
from bokeh.core.property.wrappers import PropertyValueColumnData

from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.plots import ImagePlot
from pyscivis.visualizer.standalone.file_selector import select_file_tkinter, select_file_from_list
from pyscivis.visualizer.util.themes import theme_map
from pyscivis.visualizer.extensions import FlatExtension, NestedExtension
from pyscivis.visualizer.extensions import extension_manager as ext_manager
from pyscivis.visualizer.demo import demo

from pyscivis.visualizer.plots.components.loading_div import LoadingDiv
from pyscivis.visualizer.models import Tree
from pyscivis.visualizer.dataclasses.config import Config


class MainComponent:
    """
    Main component displayed in the standalone app.
    """

    def __init__(self,
                 config: Config,
                 file: Optional[str] = None,
                 noselector: bool = False,
                 server: Optional[str] = None,
                 **kwargs: Any
                 ) -> None:
        """
        Set up file-selector and its callback, show file if necessary and initialize intro-button.

        Args:
            config: Config object.
            file: If there is a file to be displayed initially.
            noselector: If the file-selector should be hidden.
            **kwargs: Keyword arguments being swallowed.
        """
        self.config = config
        self.leaf_dict = dict()
        self.loading_div = LoadingDiv()

        ext_manager.check_extension_validity()
        if not noselector:  # create File-Selection button
            valid_extensions = ext_manager.get_supported_files()

            def sel_file_wrapper(extension_name: str):
                multiple = ext_manager.extension_allows_multiple(extension_name)
                if server is None:
                    filename = select_file_tkinter(extension_name, valid_extensions, multiple)
                    if filename:
                        self.update(filename, extension_name)
                else:
                    file_root = server
                    valid_ext_names = [filetype for alias, _, filetype in valid_extensions if alias == extension_name][0]
                    tree = select_file_from_list(file_root, valid_ext_names)

                    tree_widget = Tree(tree=tree, sizing_mode="stretch_both")
                    saved_layout = self.layout.children[0]
                    for layout in self.layout.children:
                        layout.visible = False
                    self.layout.children[0] = tree_widget

                    def select_file_and_restore_layout(attr, old, new):
                        _, filename, _ = new
                        self.layout.children[0] = saved_layout
                        for layout in self.layout.children:
                            layout.visible = True
                        self.update(filename, extension_name)
                    tree_widget.on_change("selected", select_file_and_restore_layout)

            file_input_btn = Dropdown(label="Select File", menu=[(desc, alias) for alias, desc, _ in valid_extensions])
            file_input_btn.on_click(lambda x: sel_file_wrapper(x.item))
        else:
            file_input_btn = Spacer()

        self.layout = row(column(file_input_btn, Spacer()),
                          column(Div(), sizing_mode="stretch_both", margin=(50, 50, 50, 50)),
                          column(self.loading_div.layout, width=0, height=0, name="hidden_row")
                          )
        self.layout_backup = None
        self.init_intro()
        self.theme = "default"
        self.init_theme_toggle()


        if file is not None:
            self.update(file)

    def update(self,
               path: str,
               extension_name: Optional[str] = None
               ) -> None:
        """
        Load a new file and update the filetree-widget.

        Args:
            extension_name: Name of extension to be used to handle the file, deduced from file-path if None.
            path: Path of file to be loaded.
        """
        handler: Union[FlatExtension, NestedExtension] = ext_manager.get_extension_handler(
                                                                extension_name=extension_name,
                                                                path=path,
                                                                config=self.config
                                                            )
        if isinstance(handler, NestedExtension):
            tree_list = handler.tree_struct
            self.leaf_dict = handler.data_leaves
            file_tree = Tree(tree=tree_list, theme=self.theme, width=300, height=600)
            file_tree.on_change("selected", lambda attr, old, new: self.change_plot(new, handler))
            self.layout.children[0].children[1] = file_tree
            self.layout.children[1].children[0] = Spacer()  # Space for  plot
        elif isinstance(handler, FlatExtension):
            self.layout.children[0].children[1] = Spacer()  # Delete old FileTree
            self.layout.children[1].children[0] = Spacer()
            self.change_plot(None, handler)

    def change_plot(self, leaf_header: Optional[Tuple[str, str, str]], handler) -> None:
        """
        Change the displayed plot to the one selected with the file-tree.

        Args:
            handler: ExtensionHandler used for parsing and creating the plot.
            leaf_header: (leaf_name, container_name, header)-tuple used to get data from the current file.
        """
        self.loading_div.set_text("Loading plot...")
        #self.layout.children[2].children[0] = self.loading_div.layout  # add loading overlay
        doc = curdoc()

        def calc_and_replace() -> None:  # entry-point for thread
            if leaf_header is not None:
                parent_name, own_name, header_type = leaf_header
                plot = handler.create_plot(
                    container_name=parent_name,
                    leaf_name=own_name,
                    container_type=header_type,
                    loading_div=self.loading_div
                )
            else:
                plot = handler.create_plot(loading_div=self.loading_div)

            @gen.coroutine
            def replace_plot() -> None:
                self.loading_div.hide()
                self.layout.children[1].children[0] = plot.get_layout()
            doc.add_next_tick_callback(replace_plot)  # can only modify doc like this outside of main thread

        # have to start a new thread to keep the main
        # thread idle so that document callbacks can fire
        threading.Thread(target=calc_and_replace).start()

    def init_theme_toggle(self) -> None:
        """
        Initialize the dark-mode toggle with hidden bokeh-buttons.
        """
        dark_js = CustomJS(code="""
                    const dark_styles = document.querySelectorAll(".dark-style")
                    const default_styles = document.querySelectorAll(".default-style")

                    for (let i=0; i<dark_styles.length; i++) {
                        dark_styles[i].removeAttribute("media")
                    }
                    for (let i=0; i<default_styles.length; i++) {
                        default_styles[i].setAttribute("media", "max-width: 1px")
                    }
                    $("#root").jstree("set_theme","default-dark");
                """)

        light_js = CustomJS(code="""
                    const dark_styles = document.querySelectorAll(".dark-style")
                    const default_styles = document.querySelectorAll(".default-style")

                    for (let i=0; i<dark_styles.length; i++) {
                        dark_styles[i].setAttribute("media", "max-width: 1px")
                    }
                    for (let i=0; i<default_styles.length; i++) {
                        default_styles[i].removeAttribute("media")
                    }
                    $("#root").jstree("set_theme","default");
                """)

        def theme_cb(theme):
            self.theme = theme
            curdoc().theme = theme_map[theme]

        btn_light = Button(css_classes=["btn-lightmode"], visible=False)
        btn_light.js_on_click(light_js)
        btn_light.on_click(lambda x: theme_cb("default"))

        btn_dark = Button(css_classes=["btn-darkmode"], visible=False)
        btn_dark.js_on_click(dark_js)
        btn_dark.on_click(lambda x: theme_cb("dark"))

        self.layout.select_one({'name': 'hidden_row'}).children.extend((btn_light, btn_dark))

    def init_intro(self) -> None:
        """
        Initialize the intro-button.
        """
        # The intro button is hidden and invoked from
        # the frontend by calling $(.bokeh-intro).click()
        # which happens if the frontend-only intro-button
        # is clicked
        intro_button = Button(css_classes=["bokeh-intro"], visible=False)
        intro_state = ColumnDataSource(data=dict(step=[-1]))

        intro_state.on_change("data", lambda attr, old, new: self.set_intro_step(new))
        intro_button.js_on_click(CustomJS(args=dict(cds=intro_state),
                                          code="""
                                              window.intro_state = cds
                                          """))

        self.layout.select_one({'name': 'hidden_row'}).children.append(intro_button)

    def set_intro_step(self, new_data: PropertyValueColumnData):
        """
        Start the intro or revert to the initial state, depending on the step provided.

        Args:
            new_data: Dictionary-like object containing a "step" key-value pair.
        """
        if "step" not in new_data.keys():
            raise ValueError("Unexpected CDS-attribute change detected.")
        step = new_data["step"][0]

        if step == 0 and self.layout_backup is None:
            # all the slicing shenanigans are because we need to keep the hidden intro_button in the layout
            # for some reason the server-side callback will be removed if he is not in the layout
            self.layout_backup = self.layout.children[:-1]  # saving the layouts row
            file_input_btn = Button(label="Select File", css_classes=["fileinputbtn", "invisible"])
            file_input_btn.js_on_click(CustomJS(code="setStep(4)"))
            tree_list = [
                {"id": "root", "parent": "#", "text": "Your_File.h5", "type": "file"},
                {"id": "container0", "parent": "root", "text": "dataset0", "type": "container"},
                {"id": "container1", "parent": "root", "text": "dataset1", "type": "container"},
                {"id": "container2", "parent": "root", "text": "dataset1", "type": "container"},
                {"id": "Acq0", "parent": "container0", "text": "ImportantImages", "type": "images"},
                {"id": "Acq1", "parent": "container1", "text": "MoreImportantImages", "type": "images"},
                {"id": "Acq2", "parent": "container2", "text": "MostImportantImages", "type": "images"},
            ]
            file_tree = Tree(tree=tree_list, theme=self.theme, width=300, height=600, css_classes=["invisible"])
            file_tree.js_on_change("selected", CustomJS(code="setStep(5)"))

            images = demo.get_images(640, 525)
            data = ParsedData(images,
                              dim_names=["Images", "Y", "X"],
                              dim_lengths=[3, 600, 1000],
                              dim_units=["pixel", "mm", "mm"])
            plot = ImagePlot(data, self.config.image).get_layout()
            plot.css_classes = ["main-app", "invisible"]
            self.layout.children[:-1] = \
                [
                    column(file_input_btn, file_tree),
                    column(plot, margin=(50, 50, 50, 50))
                 ]

        if step == -1:
            self.layout.children[:-1] = self.layout_backup
            self.layout_backup = None



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pyscivis",
    version="0.9.1",
    author="Jannik Meinecke",
    #author_email="TODO",
    description="A visualization tool for ISMRMRD files. Can be used standalone or embedded into Jupyter Notebooks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/chi-imrt/pyscivis/pyscivis",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'bokeh==2.3.1',
        'tornado==6.1',
        'ismrmrd==1.8.0',
        'Pillow==8.1.2',
        'numpy==1.21.4',
        'numba==0.53.1',
        'attr==0.3.1',
        'desert==2020.11.18',
        'toml==0.10.2',
        'pptree==3.1',
        'defusedxml==0.7.1'
    ],
    extras_require={
        'testing': [
            'pytest==6.2.4',
            'pytest-cov',
            'pytest-dependency'
        ]
    },
    project_urls={
        "Bug Tracker": "https://gitlab.com/chi-imrt/pyscivis/pyscivis/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    entry_points={'console_scripts': ['pyscivis = pyscivis.__main__:main']},
    python_requires=">=3.8",
)



from typing import Optional

import numpy as np

from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.extensions import NestedExtension
from pyscivis.visualizer.plots import ImagePlot, TablePlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv


class ExampleNestedExtension(NestedExtension):  # pragma: nocover

    alias = "example_nested"
    multiple = True
    supported_files = "*"
    file_description = "Create Image for each filepath w/ TreeWidget"

    def create_plot(self,
                    leaf_name: str,
                    container_name: str,
                    container_type: str,
                    loading_div: Optional[LoadingDiv] = None):

        path = self.data_leaves[leaf_name]

        if container_type == "header":
            path: tuple   # path is a tuple for the header
            column_dict = dict(keys=list(), values=list())
            for index, p in enumerate(path):
                column_dict["keys"].append(f"Image{index}'s path")
                column_dict["values"].append(p)
            column_dict["keys"].append("Length of all paths combined")
            comb_length = len("".join(path))
            column_dict["values"].append(comb_length)
            return TablePlot(column_dict)
        else:  # image
            path: str
            np.random.seed(seed=len(path))  # let's make the image the same for paths with the same length
            data = np.random.rand(2, 4, 3, 34, 35)
            parsed = ParsedData(
                data=data,
                dim_names=["dim4", "dim3", "dim2", "dim1", "dim0"],
                dim_lengths=np.array(data.shape)*2,
                dim_units=["mm", "mm", "mm", "mm", "mm"],
            )
            return ImagePlot(parsed, self.config.image)

    def get_valid_leaf_type_combinations(self):
        return list(self.data_leaves.keys())

    @staticmethod
    def get_tree_structure(file_path):

        tree_structure = list()
        root_name = f"{len(file_path)} random images"
        root = {"id": "root", "parent": "#", "text": root_name, "type": "file"}
        tree_structure.append(root)
        header = {"id": "header", "parent": "root", "text": "Header", "type": "header"}
        tree_structure.append(header)
        for index, path in enumerate(file_path):
            node = {"id": path, "parent": "root", "text": str(index), "type": "images"}
            tree_structure.append(node)
        return tree_structure

    @staticmethod
    def get_data_leaves(file_path):
        data_leaves = dict()
        for path in file_path:
            data_leaves[path] = path
        data_leaves["header"] = file_path  # tuple of all paths
        return data_leaves




from typing import Callable

from bokeh.layouts import row, column
from bokeh.models import TextAreaInput, Button, CustomJS, FileInput
from pyscivis.visualizer.dataclasses.state import state_to_json, state_from_json, is_json_state, State
import base64


download_state_code = \
    """
    function doDownload(str) {
      function dataUrl(data) {
        return "data:x-text/plain;charset=utf-8," + escape(data);
      }
      var downloadLink = document.createElement("a");
      downloadLink.href = dataUrl(str);
      downloadLink.download = "state.txt";
    
      document.body.appendChild(downloadLink);
      downloadLink.click();
      document.body.removeChild(downloadLink);
    }
    doDownload(textbox.value)
    """


def StateFigure(format_state: bool,
                set_state_cb: Callable[[State], None],
                get_state_cb: Callable[[], State]):
    """
    Create a StateFigure (Textbox, {upload, download, get-state, set-state}-buttons)

    Args:
        format_state: Whether the state should be formatted in a pretty way.
        set_state_cb: Callback fired if a state change was fired by the user.
        get_state_cb: Callback fired if the user requests the current state.

    Returns:
        A bokeh LayoutDOM containing the StateFigure

    """
    # TextBox

    state_text_box = TextAreaInput(rows=14, max_length=None)
    indent = 4 if format_state else None

    # Get- and Set-state buttons
    def set_state_text(state: State) -> None:
        text = state_to_json(state, indent=indent)
        state_text_box.value = text

    get_state_btn = Button(label="Get State")
    get_state_btn.on_click(lambda x: set_state_text(get_state_cb()))
    set_state_btn = Button(label="Set State", disabled=True)
    set_state_btn.on_click(lambda x:
                           set_state_cb(state_from_json(state_text_box.value)))
    text_btn_row = row(get_state_btn, set_state_btn)

    # Download button
    download_state_btn = Button(label="Download State", disabled=True)
    download_state = CustomJS(args=dict(textbox=state_text_box),
                              code=download_state_code)
    download_state_btn.js_on_click(download_state)

    # Upload Button
    hidden_file_upload_btn = FileInput(accept=".txt", multiple=False, visible=False, css_classes=["file-input"])

    def set_text_from_b64(s_b64):
        s_bytes = base64.b64decode(s_b64)
        state_text_box.value = s_bytes.decode('utf-8')

    hidden_file_upload_btn.on_change("value", lambda attr, old, new: set_text_from_b64(new))

    upload_state_btn = Button(label="Upload State")  # using this as a visually matching "fake"-file-input
    upload_state_btn.js_on_click(CustomJS(code="document.querySelector('.file-input input').click()"))

    def on_text_change(new_text):
        disabled = not is_json_state(new_text)
        set_state_btn.disabled = disabled
        download_state_btn.disabled = disabled

    state_text_box.on_change("value", lambda attr, old, new: on_text_change(new))

    file_btn_row = row(download_state_btn, upload_state_btn, hidden_file_upload_btn)
    state_layout = column(state_text_box, text_btn_row, file_btn_row)

    return state_layout



from typing import Sequence

import attr
from numpy import ndarray


@attr.s()
class ParsedData:
    """
    A dataclass containing the data put together by a Parser.

    The data contained in objects of this class is used by plots to draw onto the canvas.
    All arguments are mandatory, e.g., "dim_lengths" cannot be set to None, even if it is equal to the shape of data.

    Args:
        data (np.ndarray): An n-dimensional numpy-array containing dimension data.
        dim_names (Sequence[str]): A sequence of strings resembling the names of the dimensions.
        dim_lengths (Sequence[float]): A sequence of floats resembling the lengths of the dimensions.
        dim_units (Sequence[str]): A sequence of strings resembling the units of the dimensions.

    Example:
        ParsedData(
            data=np.array([[0, 1], [2, 3]]),
            dim_names=["y", "x"],
            dim_lengths=[2.5, 5],
            dim_units=["mm", "mm"]
        )
    """
    data: ndarray = attr.ib()
    dim_names: Sequence[str] = attr.ib()
    dim_lengths: Sequence[float] = attr.ib()
    dim_units: Sequence[str] = attr.ib()

    def __attrs_post_init__(self):
        n_dim = self.data.ndim
        if n_dim != len(self.dim_names) or \
           n_dim != len(self.dim_lengths) or \
           n_dim != len(self.dim_units):
            raise ValueError(f"`data.ndim ({n_dim})`, `len(dim_names) ({len(self.dim_names)})`,"
                             f" `len(dim_lengths) ({len(self.dim_lengths)})`, and `len(dim_units)"
                             f" ({len(self.dim_units)})` all have to be equal.")



from typing import Any, Dict, Sequence, Callable

from bokeh.io import curdoc
from bokeh.layouts import column, row
from bokeh.models import Button, Select

from pyscivis.visualizer.models import ThrottledSlider
from pyscivis.visualizer.dataclasses.state import DimensionState, AxesState

from pyscivis.visualizer.dataclasses.handler import Dim


class DimensionControls:
    """
    Manage dimension related controls, i.e., Axes-dropdowns and Dimension-sliders.
    """

    def __init__(self,
                 dim_metadata: Dict[int, Dim],
                 dim_change_cb: Callable[[int, int], Any],
                 axes_change_cb: Callable[[int, int], Any],
                 ) -> None:
        """
        Set up objects and callbacks.

        Args:
            dim_metadata: Metadata for dimensions.
            dim_change_cb: Callback to be used if dimension-values change.
            axes_change_cb: Callback to be used if axes change.
        """
        self.notify_dim_change = dim_change_cb
        self.notify_axes_change = axes_change_cb

        self.animated_dimension = None
        self.dim_sliders = list()

        self.axis_select = DimensionSelectPair(dim_metadata, self)

        # Create a bunch of DimensionSliders
        for k, v in dim_metadata.items():
            if v.size == 1:
                self.dim_sliders.append(None)
            else:
                self.dim_sliders.append(DimensionSlider(k, v.name, v.size, self))

        # Add only valid layouts to the main layout
        layout = list()
        for slider in reversed(self.dim_sliders):
            if slider is not None:
                layout.append(slider.layout)

        self.slider_layout = column(*layout, sizing_mode="stretch_width", css_classes=["slider-column"])
        self.hide_sliders([i for i in range(len(dim_metadata))][-2:])  # Hide last two axes

        self.axis_select_layout = self.axis_select.layout

    def set_state(self, state: DimensionState) -> None:
        """
        Set the state of the objects managed by this class.

        Args:
            state: DimensionState to be set.

        See Also:
            DimensionSlider, DimensionSelectPair
        """
        # TODO: (?) could use document.hold to reduce the amount of on_changes called
        axes: AxesState = state.axes
        dimensions: Dict[str, int] = state.dimensions

        self.axis_select.set_x_without_cb(axes.x_axis)
        self.axis_select.y_select.value = axes.y_axis
        for name, val in dimensions.items():
            for slider in self.dim_sliders:  # find correct slider for dim
                if slider is None or slider.name != name:
                    continue

                if val < slider.slider.start:
                    val = slider.slider.start
                elif val > slider.slider.end:
                    val = slider.slider.end

                slider.slider.value = val

    def get_state(self) -> DimensionState:
        """
        Get the state of the objects managed by this class.

        Returns:
            A State object.

        See Also:
            DimensionSlider, DimensionSelectPair
        """
        x_axis = self.axis_select.x_select.value
        y_axis = self.axis_select.y_select.value
        axes = AxesState(x_axis=x_axis, y_axis=y_axis)

        dimensions: Dict[str, int] = dict()
        for slider in self.dim_sliders:
            if slider is None:
                continue
            dimensions[slider.name] = slider.slider.value

        state = DimensionState(axes=axes, dimensions=dimensions)
        return state

    def on_axes_change(self,
                       x_axis: int,
                       y_axis: int
                       ) -> None:
        """
        Hide sliders if their dimension is displayed and notify controller of axes-change.

        Also disables the currently animated slider if called.

        Args:
            x_axis: Index of the new x-axis.
            y_axis: Index of the new y-axis.
        """
        self.disable_active_slider()
        for slider in self.dim_sliders:
            if slider is None:
                continue
            if slider.id in [x_axis, y_axis]:
                slider.hide()
            else:
                slider.show()
        self.notify_axes_change(x_axis, y_axis)

    def hide_sliders(self, axes: Sequence[int]) -> None:
        """
        Make the specified axes invisible.

        Args:
            axes (Sequence[int]): A sequence of axis-indices.
        """
        for axis in axes:
            if self.dim_sliders[axis] is not None:
                self.dim_sliders[axis].hide()

    def toggle_slider(self, index: int) -> None:
        """
        Animate a dimension slider by indefinitely increasing its value.

        A periodic callback is registered on the bokeh document that
        increases the slider-value by 1 (wraps around the maximum).
        If this method is called while another dimension slider is being
        animated, stop that animation.

        Args:
            index: Index of the axis to be animated.
        """
        if self.animated_dimension is not None and self.animated_dimension != index:
            self.disable_active_slider()
        global callback_id

        dim_slider = self.dim_sliders[index]

        if dim_slider.play_btn.label == '►':
            self.animated_dimension = index
            dim_slider.play_btn.label = '❚❚'
            callback_id = curdoc().add_periodic_callback(dim_slider.inc_slider, 200)
        else:
            self.animated_dimension = None
            dim_slider.play_btn.label = '►'
            curdoc().remove_periodic_callback(callback_id)

    def disable_active_slider(self) -> None:
        if self.animated_dimension is None:
            return
        global callback_id
        self.dim_sliders[self.animated_dimension].play_btn.label = '►'
        curdoc().remove_periodic_callback(callback_id)
        self.animated_dimension = None


class DimensionSlider:
    """
    Wrap a slider, callback and data for identifying objects of this class.
    """

    def __init__(self,
                 index: int,
                 name: str,
                 length: int,
                 controller: DimensionControls
                 ) -> None:
        """
        Set up attributes, the slider, button and necessary callbacks.

        Args:
            index: Index for this slider.
            name: Name of this slider.
            length: Size of this slider.
            controller: Parent controlling this object.
        """

        # TODO: (?) onDragStart = disable_active_slider
        self.id = index
        self.name = name
        self.length = length
        self.slider = ThrottledSlider(base_delay=200, title=name, start=0, end=length-1,
                                      value=0, step=1, sizing_mode="stretch_width")
        self.slider.on_change('value', lambda attr, old, new: controller.notify_dim_change(index, new))

        dec_btn = Button(label='<', aspect_ratio=1, align="center")
        self.play_btn = Button(label='►', aspect_ratio=1, align="center")
        inc_btn = Button(label='>', aspect_ratio=1, align="center")

        dec_btn.on_click(self.dec_slider)
        self.play_btn.on_click(lambda: controller.toggle_slider(index))
        inc_btn.on_click(self.inc_slider)

        btn_row = row(dec_btn, self.play_btn, inc_btn)
        self.layout = row(self.slider, btn_row)

    def inc_slider(self):
        if self.slider.value == self.slider.end:
            self.slider.value = self.slider.start
        else:
            self.slider.value += 1

    def dec_slider(self):
        if self.slider.value == self.slider.start:
            self.slider.value = self.slider.end
        else:
            self.slider.value -= 1

    def show(self) -> None:
        """
        Make self visible.
        """
        self.layout.visible = True

    def hide(self) -> None:
        """
         Make self invisible.
        """
        self.layout.visible = False


class DimensionSelectPair:
    """
    Wrap two axes-selection widgets and their interaction.
    """

    def __init__(self,
                 options: Dict[int, Dim],
                 controller: DimensionControls
                 ) -> None:
        """
        Set up both selects and their callbacks.

        Args:
            options: Dictionary mapping dimension-indices to Dim metadata.
            controller: Parent controlling this object.
        """
        self.options = {v.name: k for k, v in options.items() if v.size > 1}
        self.controller = controller
        self.x_select = Select(title="X-Axis", options=sorted(self.options.keys()))
        self.y_select = Select(title="Y-Axis", options=sorted(self.options.keys()))

        self.x_select.value = list(self.options.keys())[-1]
        self.y_select.value = list(self.options.keys())[-2]

        self.x_select.on_change('value', self.on_x_change)
        self.y_select.on_change('value', self.on_y_change)

        self.layout = column(self.x_select, self.y_select, css_classes=["axes-selector"])

    def on_x_change(self,
                    attr: Any,
                    old: str,
                    new: str
                    ) -> None:
        """
        Checks if the y-select currently has the new value and swaps them if yes, then calls a callback.

        Args:
            attr: Attribute name.
            old: Former name of the x-axis.
            new: New name for the y-axis.

        Note:
            Converts the name into the according index for the callback.
        """
        if new == self.y_select.value:  # workaround to avoid unnecessary cbs
            self.set_y_without_cb(old)
        x_new = self.options[new]
        y_new = self.options[self.y_select.value]
        self.controller.on_axes_change(x_new, y_new)

    def on_y_change(self,
                    attr: Any,
                    old: str,
                    new: str
                    ) -> None:
        """
        Checks if the x-select currently has the new value and swaps them if yes, then calls a callback.

        Args:
            attr: Attribute name.
            old: Former name of the y-axis.
            new: New name for the y-axis.

        Note:
            Converts the name into the according index for the callback.
        """
        if new == self.x_select.value:  # workaround to avoid unnecessary cbs
            self.set_x_without_cb(old)
        x_new = self.options[self.x_select.value]
        y_new = self.options[new]
        self.controller.on_axes_change(x_new, y_new)

    def set_x_without_cb(self, val: str) -> None:
        """
        Set a value to the x-axis without calling the callback.

        Args:
            val: New name for the x-axis.
        """
        self._remove_x_select_cb()
        self.x_select.value = val
        self._attach_x_select_cb()

    def set_y_without_cb(self, val: str) -> None:
        """
        Set a value to the y-axis without calling the callback.

        Args:
            val: New name for the y-axis.
        """
        self._remove_y_select_cb()
        self.y_select.value = val
        self._attach_y_select_cb()

    def _remove_x_select_cb(self) -> None:
        self.x_select.remove_on_change('value', self.on_x_change)

    def _attach_x_select_cb(self) -> None:
        self.x_select.on_change('value', self.on_x_change)

    def _remove_y_select_cb(self) -> None:
        self.y_select.remove_on_change('value', self.on_y_change)

    def _attach_y_select_cb(self) -> None:
        self.y_select.on_change('value', self.on_y_change)



from abc import ABC, abstractmethod
from typing import Optional, Union, List, Dict, Any, Sequence

from pyscivis.visualizer.dataclasses.config import Config
from pyscivis.visualizer.plots import ImagePlot, ComplexPlot, TablePlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv


class FlatExtension(ABC):
    """
    Base class for all extensions that do not wish to use a file browser to navigate through data.

    Strongly recommended for single non-hierarchical files, e.g., .jpg, .png, .txt, etc.

    Also recommended, if you want to 'merge' multiple selected files into
    one piece of data and display them as one plot.
    If you want to navigate multiple files independently (using a file browser)
    use NestedExtension instead.
    """
    @property
    @abstractmethod
    def alias(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def multiple(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def supported_files(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def file_description(self):
        raise NotImplementedError()

    def __init__(self,
                 path: Union[str, Sequence[str]],
                 config: Config
                 ) -> None:
        """
        Set up class attributes.

        Args:
            path: Path of file to be loaded.
            config: Config object.
        """

        if self.multiple is True:  # want to make sure we check a boolean not a property
            if isinstance(path, str):
                path = (path, )  # compatibility for notebook path entry

        self.path = path
        self.config = config

    @abstractmethod
    def create_plot(self,
                    loading_div: Optional[LoadingDiv] = None
                    ) -> Union[ImagePlot, ComplexPlot, TablePlot]:
        """
        Create a plot by reading data directly from a non-nested file.

        Raises:
            NotImplementedError: This method is meant to be overridden.
        """
        raise NotImplementedError()


class NestedExtension(FlatExtension):
    """
    Base class for all extensions that want to create a file browser for the user to choose data from.

    Recommended for hierarchical files like 'h5'-files or if selecting
    a large amount of unrelated data-files.
    If you wish to merge all data of an h5 into one plot and
    do not need a file browser, use FlatExtension instead.
    """

    def __init__(self,
                 path: Union[str, Sequence[str]],
                 config: Config
                 ) -> None:
        """
        Set up file handler and required data structures for a ContainerExtension.

        Args:
            path: Path of the file to be loaded.
            config: Config object.
        """
        super().__init__(path, config)
        self.tree_struct = self.get_tree_structure(self.path)
        self.data_leaves = self.get_data_leaves(self.path)

    @abstractmethod
    def create_plot(self,
                    leaf_name: str,
                    container_name: str,
                    container_type: str,
                    loading_div: Optional[LoadingDiv] = None
                    ) -> Union[ImagePlot, ComplexPlot, TablePlot]:
        """
        Create a plot and return its layout.

        Args:
            leaf_name: Full ID of clicked leaf.
            container_name: Name of container with plot-data.
            container_type: Type of the container.
            loading_div: Used to notify user of progress, if None -> stdout.

        Returns:
            An ImagePlot or ComplexPlot .

        Raises:
            NotImplementedError: This method is meant to be overridden.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_valid_leaf_type_combinations(self) -> List[str]:
        """
        Infer valid name-type combinations from leaf-list.

        Raises:
            NotImplementedError: This function is meant to be overridden.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_tree_structure(file_path: str) -> List[Dict[str, str]]:
        """
        Construct a flat tree-structure to be used by the filetree-widget.

        Args:
            file_path: Path of file to be parsed into a tree structure.

        Raises:
            NotImplementedError: This method is meant to be overridden.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_data_leaves(file_path: str) -> Dict[str, Any]:
        """
        Get (container-name, container-data)-Dictionary.

        Args:
            file_path: Path of file to be parsed.

        Raises:
            NotImplementedError: This method is meant to be overridden.
        """
        raise NotImplementedError()



"""
This module was adapted from https://github.com/esa-esdl/gridtools

                        The MIT License (MIT)

Copyright (c) 2016, Brockmann Consult GmbH and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import Tuple, Any

from numba import prange, njit
import numpy as np

_EPS = 1e-10

def downsample_2d_mean_nojit(src: np.ndarray,
                       shape: Tuple[int, int],
                       fill_value: Any = np.nan
                       ) -> np.ndarray:
    """
    Downsample a 2D-numpy-array to the specified shape.

    Args:
        src: 2D-array to be downsampled.
        shape: (height, width)-shape of the downsampled array.
        fill_value: Value to be used instead infinitesimal values (< 1e-10).

    Returns:
        A downsampled numpy-array with the specified shape.

    Raises:
        ValueError: If specified shape is larger than source shape at any point.

    Notes:
        Adapted from https://github.com/esa-esdl/gridtools/blob/master/gridtools/resampling.py#L309
    """
    src_h, src_w = src.shape
    out_h, out_w = shape
    out = np.empty(shape, dtype=src.dtype)
    if src_w == out_w and src_h == out_h:
        return src

    if out_w > src_w or out_h > src_h:
        raise ValueError("invalid target size")

    scale_x = src_w / out_w
    scale_y = src_h / out_h

    for out_y in prange(out_h):
        src_yf0 = (scale_y * out_y)
        src_yf1 = (src_yf0 + scale_y)
        src_y0 = int(src_yf0)
        src_y1 = int(src_yf1)

        wy0 = 1.0 - (src_yf0 - src_y0)
        wy1 = src_yf1 - src_y1
        if wy1 < _EPS:
            wy1 = 1.0
            if src_y1 > src_y0:
                src_y1 -= 1
        for out_x in range(out_w):
            src_xf0 = (scale_x * out_x)
            src_xf1 = src_xf0 + scale_x
            src_x0 = int(src_xf0)
            src_x1 = int(src_xf1)
            wx0 = 1.0 - (src_xf0 - src_x0)
            wx1 = src_xf1 - src_x1
            if wx1 < _EPS:
                wx1 = 1.0
                if src_x1 > src_x0:
                    src_x1 -= 1
            v_sum = 0.0
            w_sum = 0.0
            for src_y in range(src_y0, src_y1 + 1):
                wy = wy0 if (src_y == src_y0) else wy1 if (src_y == src_y1) else 1.0
                for src_x in range(src_x0, src_x1 + 1):
                    wx = wx0 if (src_x == src_x0) else wx1 if (src_x == src_x1) else 1.0
                    v = src[src_y, src_x]
                    if np.isfinite(v):
                        w = wx * wy
                        v_sum += w * v
                        w_sum += w
            if w_sum < _EPS:
                out[out_y, out_x] = fill_value
            else:
                out[out_y, out_x] = v_sum / w_sum
    return out


downsample_2d_mean = njit(downsample_2d_mean_nojit, nogil=True, parallel=True)



import math
from typing import Dict, Union, List, Tuple

import numpy as np
from bokeh.models import ColumnDataSource, CustomJS, HoverTool, WheelZoomTool, ColorBar, Image as ImageGlyph, \
    PolyAnnotation, Title
from bokeh.plotting import figure

from pyscivis.visualizer.util.downsample import downsample_2d_mean
from pyscivis.visualizer.dataclasses.config import ImageConfig

dims_jscode = """
    var update_dims = function () {
        var new_data = {
            height: [plot.height],
            width: [plot.width],
            x_start: [plot.x_range.start],
            y_start: [plot.y_range.start],
            x_end: [plot.x_range.end],
            y_end: [plot.y_range.end]
        };
        dims.data = new_data;
    };
    var throttle = window.throttle
    if (typeof throttle != 'undefined' && throttle != null) {
        clearTimeout(throttle);
    }
    
    window.throttle = setTimeout(update_dims, %i, "replace");
    """


class GriddableImage:
    """
    Takes care of keeping the displayed image up to date, regridding/downsampling and sizing it.
    """

    def __init__(self,
                 image_data: np.ndarray,
                 dw: float,
                 dh: float,
                 color_bars: Dict[str, ColorBar],
                 initial_cbar: Union[int, str],
                 use_rgba: bool,
                 config: ImageConfig,
                 title: str = ""
                 ) -> None:
        """
        Set up config attributed, callbacks, and the image including tools.

        Args:
            image_data: 2-dimensional numpy-array containing image data.
            dw: Length of the x-axis.
            dh: Length of the y-axis.
            color_bars: Dictionary mapping ColorBar names to ColorBar objects.
            initial_cbar: Either the index (in color_bars) or the name of the inital ColorBar.
            use_rgba: Whether to use an 'image_rgba'-glyph; disabled color-bar/-mapper logic.
            config: A ImageConfig object.
            title: Optional title displayed above the main plot.
        """
        self.PREF_HEIGHT = config.pref_height
        self.MIN_HEIGHT = config.min_height
        self.MAX_HEIGHT = config.max_height
        self.MIN_WIDTH = config.min_width
        self.MAX_WIDTH = config.max_width
        self.MIN_BORDER = config.border

        self.ds_downsample_images_threshold = config.ds_downsample_images_threshold
        self.ds_threshold = config.ds_threshold
        self.downsample_flag = False

        self.image_data: np.ndarray = np.array([])
        self.x_span: float = 0
        self.y_span: float = 0
        self.x_ratio: float = 0
        self.y_ratio: float = 0

        self.image_source = ColumnDataSource(data=dict(image=[], x=[], y=[], dw=[], dh=[]))
        self.client_dims = ColumnDataSource(data=dict(width=[], height=[],
                                                      x_start=[], x_end=[],
                                                      y_start=[], y_end=[]))

        self.client_dims.on_change('data', lambda attr, old, new: self.regrid_image(self.ds_threshold))

        self.plot = figure(tools=["pan", "reset"], toolbar_location="above",
                           x_range=(0, 1), y_range=(0, 1),
                           min_border_left=self.MIN_BORDER, min_border_bottom=self.MIN_BORDER,
                           min_border_top=self.MIN_BORDER, min_border_right=self.MIN_BORDER * 2,
                           css_classes=["image-figure"])
        self.plot.add_layout(Title(text=""), "above")
        self.plot.add_layout(Title(text=title, align="center"), "above")
        self.plot.toolbar.logo = None

        ds_indicator = PolyAnnotation(
            fill_color="red",
            fill_alpha=0.4,
            xs=[10, 5, 5, 10, 15, 20, 20, 15],
            ys=[5, 10, 15, 20, 20, 15, 10, 5],
            xs_units='screen',
            ys_units='screen',
            visible=False,
            name="ds_indicator"
        )
        self.plot.add_layout(ds_indicator)

        # TODO: update this in bokeh 2.4 with range_change event
        range_js_cb = dims_jscode % config.ds_throttle
        self.plot.x_range.js_on_change('start', CustomJS(code=range_js_cb,
                                                         args=dict(plot=self.plot, dims=self.client_dims)))
        self.plot.y_range.js_on_change('start', CustomJS(code=range_js_cb,
                                                         args=dict(plot=self.plot, dims=self.client_dims)))
        self.plot.x_range.js_on_change('end', CustomJS(code=range_js_cb,
                                                       args=dict(plot=self.plot, dims=self.client_dims)))
        self.plot.y_range.js_on_change('end', CustomJS(code=range_js_cb,
                                                       args=dict(plot=self.plot, dims=self.client_dims)))

        self.img_glyphs: List[ImageGlyph] = list()
        if not use_rgba:
            if isinstance(initial_cbar, int):
                initial_cbar: str = list(color_bars.keys())[initial_cbar]  # Grab key name if initial_cbar is an int

            for name, c_bar in color_bars.items():
                img = self.plot.image(source=self.image_source, name='image',
                                      image='image', x='x', y='y', dw='dw', dh='dh',
                                      color_mapper=c_bar.color_mapper)
                self.img_glyphs.append(img)
                if name != initial_cbar:
                    img.visible = False
                    c_bar.visible = False
                self.plot.add_layout(c_bar, "right")
            self.color_bars = color_bars
        else:
            img = self.plot.image_rgba(source=self.image_source, name='image',
                                       image='image', x='x', y='y', dw='dw', dh='dh')
            self.img_glyphs.append(img)

        wz_tool = WheelZoomTool(zoom_on_axis=False, maintain_focus=False)
        h_tool = HoverTool(tooltips="@image ($x{%i}, $y{%i})",
                           formatters={"$x": "printf", "$y": "printf"},
                           renderers=[glyph for glyph in self.img_glyphs])
        self.plot.add_tools(wz_tool, h_tool)

        # this will not start the js_logic because it has not been loaded into the browser by now
        self.change_image(image_data, dw, dh)

        data = dict(width=[self.plot.width], height=[self.plot.height],
                    x_start=[0], x_end=[self.x_span],
                    y_start=[0], y_end=[self.y_span])
        self.client_dims.data = data  # so we take matters in our own hands and cheat a bit here

    def change_image(self,
                     image_data: np.ndarray,
                     dw: float,
                     dh: float
                     ) -> None:
        """
        Completely change image, including its axes and resetting the bounds.

        Args:
            image_data: 2-dimensional numpy-array containing image data.
            dw: Length of the x-axis.
            dh: Length of the y-axis.
        """
        self._set_reset_ranges(dw, dh)
        self.image_data = image_data

        if image_data.size > self.ds_downsample_images_threshold**2:
            self.downsample_flag = True
        else:
            self.downsample_flag = False
            data = dict(image=[image_data], x=[0], y=[0], dw=[dw], dh=[dh])
            self.image_source.stream(data, rollover=True)

        self.x_span = dw
        self.y_span = dh
        self.x_ratio = self.x_span / self.image_data.shape[1]
        self.y_ratio = self.y_span / self.image_data.shape[0]
        width, height = self._calc_width_height(self.x_span, self.y_span)

        # This is necessary to only update the plot and ranges if they would
        # actually change, since updating the plot and ranges with old values
        # would not cause a re-grid to happen (because no change was detected)
        if self.plot.width != width or self.plot.height != height or \
                self.plot.x_range.end != dw or self.plot.y_range.end != dh or \
                self.plot.x_range.start != 0 or self.plot.y_range.start != 0:
            self.plot.update(width=width, height=height)
            self.plot.x_range.update(start=0, end=self.x_span,
                                     bounds=(0, self.x_span))  # this will trigger update image
            self.plot.y_range.update(start=0, end=self.y_span, bounds=(0, self.y_span))

        # if old values are actually the same (for fully zoomed out images with same {x, y}-size)
        else:
            self.regrid_image(self.ds_threshold)

    def update_image(self, image_data: np.ndarray) -> None:
        """
        Only update the current frame, keep axes and bounds (think next image in image-series).

        Args:
            image_data: 2-dimensional numpy-array containing image data.
        """
        self.image_data = image_data
        if not self.downsample_flag:
            self.image_source.data.update(image=[image_data])
        self.regrid_image(self.ds_threshold)

    def regrid_image(self, threshold: int) -> None:
        """
        Regrid the image depending on the current zoom.

        This function checks if the current zoom-level of an image has more values than we want to display,
        as specified by the threshold².
        If there is too many values, we grid the data and aggregate(mean) just as much as we need to make it fit
        into the threshold bounds.
        If the amount of values is fine, e.g., zoomed in or small image, it will be displayed as is.

        Args:
            threshold: Threshold-value signifying up to what point to downsample and when to use the original image.

        Notes:
            As of bokeh-2.3.2 `threshold` should not be larger than 512 (512*512 allowed datapoints)
                because the browser will struggle with drawing that amount of glyphs onto a canvas.
        """

        if not self.downsample_flag:
            # downsampling for this image was disabled because the size
            # is smaller than ds_downsample_images_threshold**2
            return

        dims_data = self.client_dims.data
        if not dims_data['width'] or not dims_data['height']:
            return
        # Note: In the following code ss_{...} variables refer to screen-space
        #                             ds_{...} variables refer to data-space

        # using max&min to make sure the ranges are in bounds
        # (bokeh tends to allow -1.2e^12 even though the bounds are at 0)
        ss_x_start = max(dims_data['x_start'][0], 0)
        ss_y_start = max(dims_data['y_start'][0], 0)
        ss_x_end = min(dims_data['x_end'][0], self.x_span)
        ss_y_end = min(dims_data['y_end'][0], self.y_span)

        # clean up bokeh-noise,
        # e.g., fully zoomed out the x-range-start can be at 1.2e^12 (should be 0)
        ss_x_start = round(ss_x_start, 6)
        ss_y_start = round(ss_y_start, 6)
        ss_x_end = round(ss_x_end, 6)
        ss_y_end = round(ss_y_end, 6)

        # calculated padded (by max 1 element on each side)
        # data-space bounds for new image
        # padding is necessary to avoid gaps at the borders
        ds_x_start = int(ss_x_start / self.x_ratio)  # int interchangeable with math.floor
        ds_x_end = math.ceil(ss_x_end / self.x_ratio)
        ds_y_start = int(ss_y_start / self.y_ratio)
        ds_y_end = math.ceil(ss_y_end / self.y_ratio)

        # calculate offset for x- and y-start of the new image
        # necessary to avoid "jumpy" behaviour when panning
        # (without the offset the entirety of a pixel would always
        #  be in the screen, this allows partial pixels)
        ss_x_offset = ss_x_start % self.x_ratio
        ss_y_offset = ss_y_start % self.y_ratio
        ss_x_start -= ss_x_offset
        ss_y_start -= ss_y_offset

        # crop image_data to only contain (padded) data for the
        # current zoom state
        new_img = self.image_data[ds_y_start:ds_y_end,
                                  ds_x_start:ds_x_end]

        # data space spans
        ds_span_x_r = (ds_x_end - ds_x_start)
        ds_span_y_r = (ds_y_end - ds_y_start)
        # screen space spans
        ss_span_x_r = ds_span_x_r * self.x_ratio
        ss_span_y_r = ds_span_y_r * self.y_ratio

        data_points_on_x_axis = ds_span_x_r
        data_points_on_y_axis = ds_span_y_r
        # check if the cropped image has more data than we are allowed to display
        if data_points_on_x_axis * data_points_on_y_axis <= threshold ** 2:
            self.plot.select_one({'name': 'ds_indicator'}).visible = False
        else:
            self.plot.select_one({'name': 'ds_indicator'}).visible = True

            # frontend plot-size
            width = dims_data['width'][0]
            height = dims_data['height'][0]

            # calculate dimensions of down-sampled image

            # safeguard to ensure width+height are always >=1
            width = max(width, 1)
            height = max(height, 1)
            # capping size at threshold because that's the point of the threshold
            # capping size at data-points-amount because the down-sampled image cannot
            #   be bigger than the original (that would be up-sampling)
            width = min(width, threshold, data_points_on_x_axis)
            height = min(height, threshold, data_points_on_y_axis)

            new_img = downsample_2d_mean(new_img, (height, width), np.nan)

        new_data = dict(image=[new_img], x=[ss_x_start], y=[ss_y_start], dw=[ss_span_x_r], dh=[ss_span_y_r])
        self.image_source.stream(new_data, rollover=True)

    def _calc_width_height(self,
                           sx: float,
                           sy: float
                           ) -> Tuple[float, float]:
        """
        Calculate the width and height of the image-plot.

        It is tried to keep the original x/y-ratio. However, if any
        of the {MIN, MAX}x{WIDTH, HEIGHT} constraints would be violated,
        the ratio is changed to fit into these constraints.

        Finally, a margin to accommodate other image-plot elements,
        like the ColorBar are added (without the margin they would
        change the ratio),

        Args:
            sx: Length of the x-axis.
            sy: Length of the y-axis.

        Returns:
            A (width, height)-tuple.
        """
        ratio = sx / sy

        min_allowed_ratio = self.MIN_WIDTH / self.MAX_HEIGHT
        max_allowed_ratio = self.MAX_WIDTH / self.MIN_HEIGHT

        if ratio <= min_allowed_ratio:  # rect is more "portrait" than bounds allow
            height = self.MAX_HEIGHT
            width = self.MIN_WIDTH

        elif ratio >= max_allowed_ratio:  # rect is more "landscape" than bounds allow
            height = self.MIN_HEIGHT
            width = self.MAX_WIDTH

        else:  # goal ratio is displayable without violating min/max bounds
            height = self.PREF_HEIGHT
            width = int(height * ratio)
            # if height was guessed too wrong we correct the bounds
            if width > self.MAX_WIDTH:
                width = self.MAX_WIDTH
                height = int(self.MAX_WIDTH / ratio)
            elif width < self.MIN_WIDTH:
                width = self.MIN_WIDTH
                height = int(self.MIN_WIDTH / ratio)

        #  Add borders to accommodate outer layout parts
        return width + 3 * self.MIN_BORDER, height + 2 * self.MIN_BORDER

    def _set_reset_ranges(self, x_end, y_end):
        """
        Reset the reset range of the plot's {x, y}-range. Necessary to make sure the reset-tool works properly.
        """
        self.plot.x_range.update(reset_start=0, reset_end=x_end)
        self.plot.y_range.update(reset_start=0, reset_end=y_end)
